@@ -1,11 +1,11 @@
 ;;; beframe.el --- Isolate buffers per frame -*- lexical-binding: t -*-
 
-;; Copyright (C) 2023-2024  Free Software Foundation, Inc.
+;; Copyright (C) 2023-2025  Free Software Foundation, Inc.
 
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
 ;; URL: https://github.com/protesilaos/beframe
-;; Version: 1.2.1
+;; Version: 1.3.0
 ;; Package-Requires: ((emacs "28.1"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -271,7 +271,7 @@ it must satisfy `framep'."
 PROMPT, DEF, REQUIRE-MATCH, and PREDICATE are the same as
 `read-buffer'.  The PREDICATE is ignored, however, to apply the
 per-frame filter."
-  (completing-read prompt
+  (completing-read (format "[Beframed] %s" prompt)
                    (beframe-buffer-names)
                    #'beframe--read-buffer-p
                    require-match
@@ -360,16 +360,11 @@ Also see the other Beframe commands:
 BUFFER is selected with completion among a list of buffers that
 belong to FRAME.
 
-Either bind this command to a key as an alternative to
-`switch-to-buffer', or enable the minor mode
-`beframe-mode' which makes all buffer prompts limit the
-candidates to those that belong to the selected frame.
-
 Also see `beframe-switch-buffer'.
 
-Raising and then selecting FRAME does not depend solely on Emacs.
-The window manager must permit such an operation.  See bug#61319:
-<https://debbugs.gnu.org/cgi/bugreport.cgi?bug=61319>."
+Note that raising and then selecting FRAME does not depend solely on
+Emacs.  The window manager must permit such an operation.  See
+bug#61319: <https://debbugs.gnu.org/cgi/bugreport.cgi?bug=61319>."
   (interactive
    (let ((obj (beframe--frame-object (beframe--frame-prompt))))
      (list obj (beframe--buffer-prompt obj))))
@@ -386,7 +381,7 @@ This is a simplified variant of `list-buffers-noselect'."
   (let* ((frame (if (framep frame) frame (selected-frame)))
          (name (frame-parameter frame 'name))
          (old-buf (current-buffer))
-         (buf (get-buffer-create (format "*Buffer List for %s*" name)))
+         (buf (get-buffer-create (format-message "*Buffer List for `%s' frame*" name)))
          (buffer-list (beframe-buffer-list frame :sort sort)))
     (with-current-buffer buf
       (Buffer-menu-mode)
@@ -430,7 +425,7 @@ Also see the other Beframe commands:
         (cons list-1 list-2)
       (cons list-2 list-1))))
 
-(defun beframe--modify-buffer-list (operation buffers)
+(defun beframe--modify-buffer-list (operation buffers &optional no-message)
   "Perform OPERATION to modify the current frame buffer list.
 
 OPERATION is a keyword to :assume or :unassume.  To assume is to include
@@ -440,7 +435,10 @@ buffer list.
 BUFFERS is a list of buffer objects to be added or removed from the
 current frame buffer list.  If BUFFERS satisfies `framep', then the list
 of buffers is that of the corresponding frame object (per
-`beframe--get-buffers')."
+`beframe--get-buffers').
+
+With optional NO-MESSAGE, do not produce a message reporting on the
+operation."
   (pcase-let* ((frame-buffers (beframe--get-buffers))
                (new-buffers (if (framep buffers)
                                 (beframe--get-buffers buffers)
@@ -455,17 +453,19 @@ of buffers is that of the corresponding frame object (per
                                frame-buffers)
                               "Unassumed"))
                   (_ (error "`%s' is an unknown operation to modify frame buffers" operation)))))
-    (if-let ((lists (beframe--get-longest-list-first frame-buffers consolidated-buffers))
-             (difference (seq-difference
-                          (mapcar #'buffer-name (car lists))
-                          (mapcar #'buffer-name (cdr lists)))))
+    (if-let* ((lists (beframe--get-longest-list-first frame-buffers consolidated-buffers))
+              (difference (seq-difference
+                           (mapcar #'buffer-name (car lists))
+                           (mapcar #'buffer-name (cdr lists)))))
         (progn
           (modify-frame-parameters nil `((buffer-list . ,consolidated-buffers)))
-          (message "%s %s buffers: %s"
-                   (propertize action 'face 'error)
-                   (propertize (format "%s" (length difference)) 'face 'warning)
-                   (propertize (format "%s" difference) 'face 'success)))
-      (message "No change to the frame's buffer list"))))
+          (unless no-message
+            (message "%s current frame %s buffers: %s"
+                     (propertize action 'face 'error)
+                     (propertize (format "%s" (length difference)) 'face 'warning)
+                     (propertize (format "%s" difference) 'face 'success))))
+      (unless no-message
+        (message "No change to the frame's buffer list")))))
 
 ;;;###autoload
 (defun beframe-assume-frame-buffers (frame)
@@ -594,11 +594,11 @@ Also see the other Beframe commands:
        (if arg
            "Buffer names matching REGEXP in the name or major mode"
          "Buffer names matching REGEXP in the name")))))
-  (if-let ((buffers (beframe--get-buffers (list regexp match-mode-names :no-internal-buffers))))
+  (if-let* ((buffers (beframe--get-buffers (list regexp match-mode-names :no-internal-buffers))))
       (beframe--modify-buffer-list :assume buffers)
     (user-error "No buffers match `%s'" regexp)))
 
-(defalias 'beframe-assume-buffers-matching-regexp 'beframe-assume-buffers-matching-regexp-all-frames
+(defalias 'beframe-assume-buffers-matching-regexp-all-frames 'beframe-assume-buffers-matching-regexp
   "Alias for `beframe-assume-buffers-matching-regexp'.")
 
 ;;;###autoload
@@ -617,12 +617,12 @@ Also see the other Beframe commands:
        (if arg
            "Buffer names matching REGEXP in the name or major mode"
          "Buffer names matching REGEXP in the name")))))
-  (if-let ((buffers (beframe--get-buffers (list regexp match-mode-names :no-internal-buffers))))
+  (if-let* ((buffers (beframe--get-buffers (list regexp match-mode-names :no-internal-buffers))))
       (beframe--modify-buffer-list :unassume buffers)
     (user-error "No buffers match `%s'" regexp)))
 
 
-(defalias 'beframe-unassume-buffers-matching-regexp 'beframe-unassume-buffers-matching-regexp-all-frames
+(defalias 'beframe-unassume-buffers-matching-regexp-all-frames 'beframe-unassume-buffers-matching-regexp
   "Alias for `beframe-unassume-buffers-matching-regexp'.")
 
 (define-obsolete-function-alias
@@ -693,7 +693,7 @@ Also see the other Beframe commands:
            "Delete buffers matching REGEXP in the name or major mode"
          "Delete buffers matching REGEXP in the name"))
       arg)))
-  (if-let ((buffers (beframe--get-buffers (list regexp match-mode-names :no-internal-buffers))))
+  (if-let* ((buffers (beframe--get-buffers (list regexp match-mode-names :no-internal-buffers))))
       (when (or beframe-kill-buffers-no-confirm
                 (y-or-n-p (format "Kill %d buffers matching `%s'?" (length buffers) regexp)))
         (mapc #'kill-buffer buffers))
@@ -714,7 +714,10 @@ Meant to be assigned to a prefix key, like this:
 (define-prefix-command 'beframe-prefix-map)
 
 (define-key beframe-prefix-map (kbd "b") #'beframe-switch-buffer)
+(define-key beframe-prefix-map (kbd "B") #'beframe-switch-buffer-in-frame)
 (define-key beframe-prefix-map (kbd "m") #'beframe-buffer-menu)
+(define-key beframe-prefix-map (kbd "r") #'beframe-rename-current-frame)
+(define-key beframe-prefix-map (kbd "R") #'beframe-rename-frame)
 (define-key beframe-prefix-map (kbd "k") #'beframe-kill-buffers-matching-regexp)
 (define-key beframe-prefix-map (kbd "a f") #'beframe-assume-frame-buffers-selectively)
 (define-key beframe-prefix-map (kbd "a F") #'beframe-assume-frame-buffers)
@@ -748,7 +751,7 @@ Also see the variable `beframe-prefix-map'."
   "Create scratch buffer in `initial-major-mode' for FRAME."
   (when beframe-create-frame-scratch-buffer
     (let* ((name (frame-parameter frame 'name))
-           (buf (get-buffer-create (format "*scratch for %s*" name))))
+           (buf (get-buffer-create (format-message "*scratch for frame `%s'*" name))))
       (with-current-buffer buf
         (funcall initial-major-mode)
         (when (and (zerop (buffer-size))
@@ -760,7 +763,7 @@ Also see the variable `beframe-prefix-map'."
                                (null frame))
                       (kill-buffer buf)))))
       (let* ((frame-bufs (beframe-buffer-list frame))
-             (frame-bufs-with-buf (append frame-bufs (list buf))))
+             (frame-bufs-with-buf (append (list buf) frame-bufs)))
         (modify-frame-parameters
          frame
          `((buffer-list . ,frame-bufs-with-buf)))))))
@@ -777,6 +780,37 @@ Also see the variable `beframe-prefix-map'."
 
 (defvar project--list) ; from project.el
 
+(defun beframe--get-frame-names (name)
+  "Return frame names equal to NAME as a list of strings."
+  (delq nil
+        (mapcar
+         (lambda (frame)
+           (when-let* ((frame-name (frame-parameter frame 'name))
+                       ((not (string-empty-p frame-name)))
+                       ((string= frame-name name)))
+             frame-name))
+         (frame-list))))
+
+(defun beframe--generate-unique-frame-name (name)
+  "Generate a unique frame name starting with NAME.
+If NAME is unique, return it as-is.  Otherwise, append <2>, <3>, etc.
+until a unique name is found."
+  (let ((frame-names (beframe--get-frame-names name)))
+    ;; Because this happens after the frame is created, if the length
+    ;; is 1, then we do not need to uniquify the name: it should be
+    ;; unique already.  This way, we avoid the scenario where some
+    ;; command already names the frame equal to NAME and then we
+    ;; increment the number.  It happened to me frequently with
+    ;; `dired-other-frame' and related.
+    (if (or (null frame-names) (= (length frame-names) 1))
+        name
+      (let ((n 2)
+            (candidate-name name))
+        (while (beframe--get-frame-names candidate-name)
+          (setq candidate-name (format "%s<%d>" name n)
+                n (1+ n)))
+        candidate-name))))
+
 (defun beframe-infer-frame-name (frame name)
   "Infer a suitable name for FRAME with given NAME.
 See `beframe-rename-frame'."
@@ -789,18 +823,19 @@ See `beframe-rename-frame'."
                           (listp project--list)
                           (or
                            (member (list dir) project--list)
-                           (member (list (abbreviate-file-name dir)) project--list)))))
-      (cond
-       ((and name (stringp name))
-        name)
-       ((and projectp buf-name)
-        (format "%s" (file-name-nondirectory (directory-file-name dir))))
-       ((and (not (minibufferp)) file-name)
-        (format "%s %s" buf-name dir))
-       ((not (minibufferp))
-        buf-name)
-       (t
-        dir)))))
+                           (member (list (abbreviate-file-name dir)) project--list))))
+           (processed-name (cond
+                            ((and name (stringp name))
+                             name)
+                            ((and projectp buf-name)
+                             (file-name-nondirectory (directory-file-name dir)))
+                            ((and (not (minibufferp)) file-name)
+                             (format "%s %s" buf-name dir))
+                            ((not (minibufferp))
+                             buf-name)
+                            (t
+                             dir))))
+      (beframe--generate-unique-frame-name processed-name))))
 
 ;;;###autoload
 (defun beframe-rename-frame (frame &optional name)
@@ -841,6 +876,16 @@ while also reviewing `beframe-infer-frame-name'."
    frame
    (list (cons 'name (beframe-infer-frame-name frame name)))))
 
+;;;###autoload
+(defun beframe-rename-current-frame ()
+  "Convenience wrapper of `beframe-rename-frame' to rename the current frame."
+  (declare (interactive-only t))
+  (interactive)
+  (let ((frame (selected-frame)))
+    (modify-frame-parameters
+     frame
+     (list (cons 'name (beframe-infer-frame-name frame nil))))))
+
 (defun beframe-maybe-rename-frame (frame &optional name)
   "Helper function to determine if `beframe-rename-function' is called.
 FRAME and optional NAME arguments are passed to the
@@ -877,7 +922,7 @@ If FRAME is nil, use the current frame."
 
 (defun beframe-do-not-assume-last-selected-buffer (&rest _)
   "Unassume the buffer of the most recently used window from the new frame."
-  (beframe--modify-buffer-list :unassume (list (window-buffer (get-mru-window)))))
+  (beframe--modify-buffer-list :unassume (list (window-buffer (get-mru-window))) :no-message))
 
 (defun beframe--with-other-frame (&rest app)
   "Apply APP with `other-frame-prefix'.
